#!/bin/env python
# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC . Under the terms of Contract
# DE-NA0003525 with National Technology and Engineering Solutions of Sandia, LLC, the U.S. Government
# retains certain rights in this software.

import pprint
import slycat.web.client

parser = slycat.web.client.ArgumentParser("Get details of project from Slycat server.")
parser.add_argument("pid", help="The ID of the project to retrieve")
arguments = parser.parse_args()

connection = slycat.web.client.connect(arguments)
project = connection.get_project(arguments.pid)
pprint.pprint(project)



# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC. 
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# Module defining Dial-A-Cluster .tdms parser, options, and a utility for
# enumerating .tdms files.

# S. Martin
# 5/31/2023

# file manipulation
import os
import fnmatch

# arrays
import numpy as np

# TDMS error handling
class TDMSUploadError (Exception):

    # exception for TDMS upload problems
    def __init__(self, message):
        self.message = message

# add tdms parser options for any dac model
def add_options(parser):

    # parsing parameters
    parser.add_argument("--min-time-points", default=10, type=int, 
        help="Minimum number of time points per channel, integer >= 2. " +
             "Default: %(default)s.")
    parser.add_argument("--min-channels", default=2, type=int,
        help="Minimum number of channels per text, integer >= 1. " +
             "Default: %(default)s.")
    parser.add_argument("--min-num-shots", default=1, type=int,
        help="Channels must occur in at least this many shots, integer >= 0. " +
             "Use zero to indicate that channel must occur in every shot. " +
             "Default: %(default)s.")
    parser.add_argument("--num-landmarks", default=None, type=int,
        help="Number of landmarks to use, integer >= 3.  Can also use zero " +
             "to indicate use of full dataset (no landmarks).  Default: %(default)s.")
    parser.add_argument("--num-PCA-comps", default=10, type=int,
        help="Number of PCA components to use, integer >= 2.  Note --num-landmarks " + 
             "over-rides --num-PCA-comps.  Default: %(default)s.")
    parser.add_argument("--overvoltage", action="store_true",
        help="Expecting overvoltage data.")
    parser.add_argument("--sprytron", action="store_true",
        help="Expecting sprytron data.")
    parser.add_argument("--intersection", action="store_true", 
        help="Combine mismatched time steps using intersection. " +
             "Default is to combine using union.")
    parser.add_argument("--do-not-infer-channel-units", action="store_true",
        help="Do not infer channel units. Default is to infer channel units " +
             "from channel name.")
    parser.add_argument("--do-not-infer-time-units", action="store_true",
        help="Do not infer time units. Default is to assume unspecified units " +
             "are seconds.")

    # exclude .tdms files
    parser.add_argument("--exclude", nargs="+",
        help='TDMS file suffixes to exclude. ' + 
             'If you want suffixes that include spaces, use quotes, ' +
             'e.g. "suffix with space".')

    return parser

# parse tmds input parameters
def get_parms(arguments):

    # set shot type
    shot_type = 'General'
    if arguments.overvoltage:
        shot_type = 'Overvoltage'
    if arguments.sprytron:
        shot_type = 'Sprytron'

    # set union type
    union_type = "Union"
    if arguments.intersection:
        union_type = "Intersection"

    # populate parameters
    parser_parms = [arguments.min_time_points, arguments.min_channels, 
                    arguments.min_num_shots, arguments.num_landmarks,
                    arguments.num_PCA_comps,
                    shot_type, union_type, 
                    not arguments.do_not_infer_channel_units,
                    not arguments.do_not_infer_time_units]
    
    return parser_parms

# check dac tdms parser parameters
def check_parms (arguments, parms):

    check_parser_msg = []

    # first parameter is minimum number of time points
    if parms[0] < 2:
        check_parser_msg.append ("Each channel must have at least two values. " + \
            "Please use a larger value for min-time-points and try again.")
    
    # second parameter is minimum number of channels
    if parms[1] < 1:
        check_parser_msg.append ("Each test must have at least one channel. " + \
            "Please use a larger value for min-channels and try again.")
    
    # third parameter is minimum number of shots
    if parms[2] < 0:
        check_parser_msg.append("Each channel must occur in at least one channel " + \
            "(use 0 to indicate every channel).  Please use a non-negative value " + \
            "for min-num-shots and try again.")

    # fourth parameter is number of landmarks
    if parms[3] is not None:
        if not (parms[3] == 0 or parms[3] >= 3):
            check_parser_msg.append("Number of landmarks must be zero or >= 3.  Please " + \
                "provide a valid number of landmarks and try again.")

    # fifth parameter is number of PCA components
    if parms[4] < 2:
        check_parser_msg.append("Number of PCA components must be >= 2.  Please provide " + \
            "a valid number of PCA components and try again.")

    # sixth parameter is expected type
    if parms[5] != "General" and \
       parms[5] != "Overvoltage" and \
       parms[5] != "Sprytron":
        check_parser_msg.append ('Expected data type must be one of "General", ' + \
            '"Overvoltage" or "Sprytron". Please use one of those options ' + \
            'and try again.')

    # seventh parameter is union or intersection (combination of time series)
    if parms[6] != "Union" and \
       parms[6] != "Intersection":
        check_parser_msg.append ('Available methods for combining mismatched, ' + \
            'time points are "Union" and "Intersection". Please use one of those options ' + \
            'and try again.')

    # seventh and eighth parameters are boolean, so either option is OK 

    # landmarks over-rides PCA comps, adjust parms
    if arguments.num_landmarks is not None:
        parms[4] = False
    else:
        parms[4] = True
        parms[3] = arguments.num_PCA_comps

    return parms, "\n".join(check_parser_msg)

# parse test data subdirectory
def parse_run_chart (run_chart):

    # default return is None
    run_chart_id = None

    # split run_chart string on "_" to get identifiers
    run_chart_id = run_chart.split("_")

    # keep only first four identifiers, if they exist
    if len(run_chart_id) >= 4:
        run_chart_id = run_chart_id[0:4]
    else:
        run_chart_id = [None, None, None, None]
    
    return run_chart_id

# helper function for catalog_tmds_files, converts string specifying batch
# numbers into a python list, e.g. 1-3,5 -> [1,2,3,5]
def convert_batches(batch_str, log):

    # an asterisk indicates any match is OK
    if batch_str == '*': 
        return None

    batches = set()
    try:
        for part in batch_str.split(','):
            x = part.split('-')
            batches.update(range(int(x[0]), int(x[-1]) + 1))
    except:
        log('Input batches = "' + batch_str + '" is an invalid batch specifier.')
        raise ValueError('Input batches = "' + batch_str + '" is an invalid batch specifier.')

    return sorted(batches)

# get suffix from a single file
def get_suffix (file):

    # get file name and extension
    head, tail = os.path.split(file)
    ext = tail.split(".")[-1].lower()

    # is it a tdms file?
    suffix = None
    if ext == 'tdms' or ext =='tdm':

        # get suffix
        suffix = tail.split("_")[-1].split(".")[0]

    return suffix

# get file suffixes for file list
def get_suffixes (file_list):

    # save results as a set
    tdms_suffixes = set()

    # go through each file and parse out suffix
    for file in file_list:

        # add suffix to set
        tdms_suffixes.add(get_suffix(file))

    return tdms_suffixes

# exclude any suffixes not desired
def exclude_suffixes (arguments, file_list):

    # get all suffixes
    include_suffixes = get_suffixes(file_list)

    # exclude suffixes as requested
    if arguments.exclude:

        # exclude and print list of excluded suffixes
        for suffix in arguments.exclude:
            include_suffixes.discard(suffix)

    # sort suffixes
    list_suffixes = list(include_suffixes)
    list_suffixes.sort()

    return list_suffixes

# catalog tdms types and matches
def list_tdms_types (file_list, list_suffixes, run_chart_ids, part_lot_batch_sn=True):

    # screen for .tdms files according to argument parameters
    run_chart_tdms_matches = []
    run_chart_tdms_types = []
    for run_chart_file in file_list:

        # get file suffix
        suffix = get_suffix(run_chart_file)
        
        # check if we want to include that suffix
        if suffix not in list_suffixes:
            continue

        # check that .tdms file has matching part_lot_batch_sn
        if part_lot_batch_sn == True:
            if not run_chart_file.startswith(
                run_chart_ids[0] + "_" + 
                run_chart_ids[1] + "_" + 
                run_chart_ids[2] + "_" + 
                run_chart_ids[3]):
                continue

        # keep track of files and types
        run_chart_tdms_types.append(suffix)
        run_chart_tdms_matches.append(run_chart_file)

    return run_chart_tdms_types, run_chart_tdms_matches

# enumerate all tdms subdirectories with files in a given directory
def catalog_tdms_dir (arguments, log):

    # gather data directories
    root_dir = arguments.input_tdms_dir[0]
    
    # check that root dir exists
    if not os.path.isdir(root_dir):
        raise TDMSUploadError("Input data directory does not exist.  Please provide " +
            "a different directory and try again.")

    # look for subdirectories with tdms files
    possible_run_chart_dirs = set()
    for dirpath, dirnames, filenames in os.walk(root_dir):
        tdms_files = [f for f in filenames if f.lower().endswith(".tdms") or f.lower().endswith(".tdm")]
        for filename in tdms_files:
            possible_run_chart_dirs.add(dirpath)
    possible_run_chart_dirs = list(possible_run_chart_dirs)
    possible_run_chart_dirs.sort()

    # check that we found data directories
    if len(possible_run_chart_dirs) == 0:
        raise TDMSUploadError('No data subdirectories found for part number match "' + 
            part_num_match + '".')

    # go through each run chart directory
    metadata = []
    for run_chart_dir in possible_run_chart_dirs:

        # get file name (ignore path)
        base_dir = os.path.basename(run_chart_dir)

        # check for part_lot_batch_sn format
        run_chart_ids = parse_run_chart(base_dir)

        # find .tdms files in run chart directory
        run_chart_files = os.listdir(run_chart_dir)
        run_chart_tdms_files = [run_chart_file for run_chart_file in run_chart_files 
                                if run_chart_file.lower().endswith('.tdms') or
                                    run_chart_file.lower().endswith('.tdm')]
    
        # check that we have .tdms files
        if len(run_chart_tdms_files) == 0:
            log ('Skipping subdirectory "' + str(run_chart_ids[0]) + '" -- does not ' +
                    'contain any TDMS files.')
            continue

        # get all suffixes
        list_suffixes = exclude_suffixes(arguments, run_chart_tdms_files)

        # check for tdms types and matches
        run_chart_tdms_types, run_chart_tdms_matches = \
            list_tdms_types(run_chart_tdms_files, list_suffixes, run_chart_ids)

        # check that we have matching .tdms files
        if len(run_chart_tdms_matches) == 0:
            log ('Skipping subdirectory "' + str(run_chart_ids[0]) + '" -- does not ' +
                    'contain any TDMS files with file matches.')
            continue

        # store in metadata structure
        metadata.append({"part": str(run_chart_ids[0]),
                        "lot": str(run_chart_ids[1]),
                        "batch": str(run_chart_ids[2]),
                        "sn": str(run_chart_ids[3]),
                        "source": os.path.abspath(run_chart_dir),
                        "tdms_files": run_chart_tdms_matches,
                        "tdms_types": run_chart_tdms_types,
                        "include_suffixes": list_suffixes})

    return metadata

# enumerate tdms files based on command line inputs, according to directory list
def catalog_tdms_files (arguments, log):

    # check for directory list style inputs
    if arguments.input_tdms_batches is not None:

        # gather data directories for given part number
        root_dir = arguments.input_tdms_batches[0]
        part_num_match = arguments.input_tdms_batches[1]

        # convert specified batches to python list
        batches = convert_batches(arguments.input_tdms_batches[2], log)

        # check that root dir exists
        if not os.path.isdir(root_dir):
            raise TDMSUploadError("Input data directory does not exist.  Please provide " +
                "a different directory and try again.")
        
        # look for directories containing data for part number provided
        root_subdirs = os.listdir(root_dir)
        part_subdirs = fnmatch.filter(root_subdirs, part_num_match + "_*")

        # look for directories containing batches
        batch_subdirs = []
        for subdir in part_subdirs:
            subdir_nums = subdir.split('_')        
            try:
                subdir_batch = int(subdir_nums[2])
                if batches == None:
                    batch_subdirs.append(subdir)
                elif subdir_batch in batches:
                    batch_subdirs.append(subdir)
            except:
                pass
    
    # check for glob style input with part_num match
    elif arguments.input_tdms_glob is not None:

        # gather data directories for given part number
        root_dir = arguments.input_tdms_glob[0]
        part_num_match = arguments.input_tdms_glob[1]

        # check that root dir exists
        if not os.path.isdir(root_dir):
            raise TDMSUploadError("Input data directory does not exist.  Please provide " +
                "a different directory and try again.")

        # look for directories containing data for part number provided
        root_subdirs = os.listdir(root_dir)
        batch_subdirs = fnmatch.filter(root_subdirs, part_num_match)
        
    # check for directory input where subdirectories are run charts
    elif arguments.input_tdms_dir is not None:
        return catalog_tdms_dir (arguments, log)

    # are there any subdirectories?
    if len(batch_subdirs) == 0:
        raise TDMSUploadError('Could not find any subdirectories for part, number matching "' +
              part_num_match + '".')

    # put subdirectories in order
    batch_subdirs.sort()

    # look through each subdirectory for run chart data
    metadata = []
    for subdir in batch_subdirs:

        # look for test data directory
        test_data_dir = os.path.join(root_dir, subdir, 'Test Data')

        # skip if run_chart_dir is not an actual directory
        if not os.path.isdir(test_data_dir):
            log('Skipping "' + test_data_dir + '" because it''s not a directory.')
            continue

        # each of the subdirectories of run_chart_dir will be a row in the 
        # metadata table for the run chart model
        possible_run_chart_dirs = os.listdir(test_data_dir)

        # sort according to data ids
        possible_run_chart_dirs.sort()

        # check that we found data directories
        if len(possible_run_chart_dirs) == 0:
            raise TDMSUploadError('No data subdirectories found for part number match "' + 
                  part_num_match + '".')

        # check that data directories conform to expected format
        for run_chart_dir in possible_run_chart_dirs:

            # check that directory is expected format for a run chart
            run_chart_ids = parse_run_chart(run_chart_dir)

            # if it's not a run chart then skip it
            if run_chart_ids[0] is None:
                log('Skipping subdirectory "' + run_chart_dir + 
                    '" because it does not conform to "part_lot_batch_serial" string format.')
                continue
        
            # find .tdms files in run chart directory
            run_chart_files = os.listdir(os.path.join(test_data_dir, run_chart_dir))
            run_chart_tdms_files = [run_chart_file for run_chart_file in run_chart_files 
                                    if run_chart_file.lower().endswith('.tdms') or
                                       run_chart_file.lower().endswith('.tdm')]
        
            # check that we have .tdms files
            if len(run_chart_tdms_files) == 0:
                log ('Skipping subdirectory "' + str(run_chart_ids[0]) + '" -- does not ' +
                      'contain any TDMS files.')
                continue

            # get all suffixes
            list_suffixes = exclude_suffixes(arguments, run_chart_tdms_files)

            # check for tdms types and matches
            run_chart_tdms_types, run_chart_tdms_matches = \
                list_tdms_types(run_chart_tdms_files, list_suffixes, run_chart_ids)

            # check that we have matching .tdms files
            if len(run_chart_tdms_matches) == 0:
                log ('Skipping subdirectory "' + run_chart_ids[0] + '" -- does not ' +
                     'contain any TDMS files with file matches.')
                continue

            # store in metadata structure
            metadata.append({"part": run_chart_ids[0],
                             "lot": run_chart_ids[1],
                             "batch": run_chart_ids[2],
                             "sn": run_chart_ids[3],
                             "source": os.path.join(os.path.abspath(test_data_dir), 
                                                    run_chart_dir),
                             "tdms_files": run_chart_tdms_matches,
                             "tdms_types": run_chart_tdms_types,
                             "include_suffixes": list_suffixes})
            
    # check that files were found
    if metadata == []:
        raise TDMSUploadError("No TDMS files matching selection criterion were found.")

    return metadata
    


# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC. 
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# Creates a batch of tdms files.

# S. Martin
# 5/31/2023

# connection to Slycat server
import slycat.web.client

# dac tdms parser options
import slycat.web.client.dac_tdms_util as dac_tdms_util
from slycat.web.client.dac_tdms_util import TDMSUploadError

# dac tdms upload
import slycat.web.client.dac_tdms as dac_tdms

# logger
import logging

# file manipulation
import os

# sets up logging to screen or file, depending on user input
def setup_logging(log_file=None):

    # set up log file, or print to screen (default)
    if log_file:

        log = logging.getLogger()
        log.setLevel(logging.INFO)

        # set up log file
        log.addHandler(logging.FileHandler(log_file))

        # restart log, in case log file already exists
        with open(log_file, 'w'):
            pass

        # log to file and screen
        def logger(msg):
            log.log(logging.INFO, msg)
            print(msg)

    else:

        # log to screen only
        def logger(msg):
            print(msg)

    return logger

# find batches
def catalog_batches(arguments, log):

    # read all meta data before further run chart specific filtering
    metadata = dac_tdms_util.catalog_tdms_files(arguments, log)

    # get batches discovered (we will create a model for each batch)
    batches = []
    part_numbers = []
    lot_numbers = []
    for tdms_file in metadata:

        # collate part, lot, batch information
        if tdms_file['batch'] not in batches:
            batches.append(tdms_file['batch'])
        if tdms_file['part'] not in part_numbers:
            part_numbers.append(tdms_file['part'])
        if tdms_file['lot'] not in lot_numbers:
            lot_numbers.append(tdms_file['lot'])

    batches.sort()

    # check that part number remains constant
    if len(part_numbers) > 1:
        log('Cannot have more than one part number for batch uploads.')
        raise TDMSUploadError('Cannot have more than one part number for batch uploads.')
    part_number = part_numbers[0]

    # check that lot number remains constant
    if len(lot_numbers) > 1:
        log('Cannot have more than one lot number for batch uploads.')
        raise TDMSUploadError('Cannot have more than one lot number for batch uploads.')
    lot_number = lot_numbers[0]
    
    # create a dictionary of files in each batch
    batch_data = {}
    for batch in batches:
        data_inds = []
        for data_ind in range(len(metadata)):
            if metadata[data_ind]['batch'] == batch:
                data_inds.append(data_ind)
        batch_data[batch] = data_inds
    
    return metadata, part_number, lot_number, batches, batch_data

# create model batches
def create_models(arguments):

    # can't have both overvoltage and sprytron
    if arguments.overvoltage and arguments.sprytron:
        raise TDMSUploadError("Can't use both overvoltage and sprytron options " +
              "together. Please select one or the other and try again.")

    # set up logging, always to screen, to file if requested
    log = setup_logging(arguments.log_file)

    # organize tdms files
    metadata, part_number, lot_number, batches, batch_data = \
        catalog_batches(arguments, log)
    
    # alter agruments for dac tdms
    del arguments.input_tdms_batches
    del arguments.input_tdms_glob
    del arguments.log_file

    for i in range(len(batches)):

        # keep track of model we are uploading
        log('Processing DAC Batch ' + batches[i] + ': (' + str(i + 1) + '/' + str(len(batches)) + ').')

        # add batch files, tally suffixes used
        batch_files = []
        batch_suffixes = set()
        for j in batch_data[batches[i]]:
            for file in metadata[j]['tdms_files']:
                batch_files.append(os.path.join(metadata[j]['source'], file))
            for suffix in metadata[j]['include_suffixes']:
                batch_suffixes.add(suffix)
        arguments.files = batch_files

        # sort suffixes
        batch_suffixes = list(batch_suffixes)
        batch_suffixes.sort()

        # keep track of what we are including/excluding
        if arguments.exclude:
            log("Excluding TDMS file suffixes:")
            for suffix in arguments.exclude:
                log("\t%s" % suffix)

        log("Including TDMS file suffixes:")
        for suffix in batch_suffixes:
            log("\t%s" % suffix)

        # change project name, model name, description to describe batch
        arguments.model_description = str(batch_suffixes)
        arguments.project_name = str(part_number)
        arguments.model_name = str(part_number) + '_' + str(lot_number) + '_' + batches[i]

        # run dac
        dac_tdms.create_model(arguments, log)

# set up argument parser
def parser ():

    # provide additional command line arguments for TDMS files
    parser = slycat.web.client.ArgumentParser(description=
        "Creates a batch of Dial-A-Cluster models using specified .tdms files.")

    # input consists of tdms batch format or tdms file
    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument("--input-tdms-glob", nargs=2,
        help="Use tdms directory glob format as input, first argument is directroy, organized by " +
             "part and lot as on the production server; second argument is part_num_match, speficied with " +
             'a unix glob using wildcards, for example "XXXXXX*", or "XXXXXX_XX*". ' + 
             'Note you should use "" in Unix to pass wildcards. Produces one run chart per directory.')

    # input tdms batches
    group.add_argument("--input-tdms-batches", nargs=3, 
        help="Use tdms batch format as input, first argument is directory, organized by part and lot " +
             'as on the production server; second argument is part_lot, e.g. "XXXXXX_XX", ' +
             'where part and lot numbers should be constant; and third argument is batches, ' +
             'e.g. "1,3,4-6,11-24".  Use "*" to designate all batches. Produces ' + 
             "one run chart per directory.")

    # optional flag to log results to file
    parser.add_argument('--log-file', default=None,
        help="Optional file to log results of TDMS uploads.  Note that if a model " +
             "fails to upload, the script will not terminate, but will carry on " + 
             "trying to upload the remaaining data.  Log file will be overwritten " +
             "if it already exists.")
    
    # model and project names/descriptions
    parser.add_argument("--marking", default="cui", 
        help="Marking type.  Default: %(default)s")
    parser.add_argument("--model-description", default="", 
        help="New model description.  Default: %(default)s")
    parser.add_argument("--model-name", default="TDMS DAC Model", 
        help="New model name.  Default: %(default)s")
    parser.add_argument("--project-description", default="", 
        help="New project description.  Default: %(default)s")
    parser.add_argument("--project-name", default="TDMS DAC Models", 
        help="New project name.  Default: %(default)s")
    
    # add tmds options
    parser = dac_tdms_util.add_options (parser)

    return parser

# dac generate batch entry point
def main():

    # set up argument parser
    batch_parser = parser()  

    # get user arguments
    arguments = batch_parser.parse_args()

    # create models
    create_models(arguments)

# command line version
if __name__ == "__main__":

    main()


# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC . Under the terms of Contract
# DE-NA0003525 with National Technology and Engineering Solutions of Sandia, LLC, the U.S. Government
# retains certain rights in this software.

import pandas as pd
import numpy 
from io import StringIO
import cherrypy

# common csv parsing code used by both slycat-web-server and slycat-web-client
# does not call slycat.web.server directly, instead returns any error messages
# error messages are returned as a list of {'type': 'warning', 'message': 'message'}
# use file_name = True to pass a file name instead of a string with file contents
def parse_file(file, file_name=False):
    """
    parses out a csv file into numpy array by column (data), the dimension meta data(dimensions),
    and sets attributes (attributes)
    :param file: csv file to be parsed
    :returns: attributes, dimensions, data, error_messages
    """
    
    # initial attributes, dimensions, and data
    # empty, but existing to avoid crashing slycat
    dimensions = [dict(name="row", type="int64", begin=0, end=0)]
    attributes = [dict(name="None", type="float64")]
    data = [numpy.zeros(0)]

    # keep track of errors
    csv_read_error = []

    # load input file as pandas dataframe
    try:
        if file_name:
            df = pd.read_csv(file)
        else:
            
            # convert to string, if not already
            # uses latin-1 the most generic encoding
            # (presumably it already failed to convert to utf-8)
            if isinstance(file, bytes):
                file = file.decode('latin-1')    

            df = pd.read_csv(StringIO(file))

    # return empty values if couldn't read file
    except Exception as e:
        csv_read_error.append({'type': 'error', 'message': 'Could not read .csv file.\n' +
                               'Pandas exception: "' + str(e) + '".'})
        return attributes, dimensions, data, csv_read_error 

    # check for at least two rows
    if df.shape[0] < 2:
        csv_read_error.append({'type': 'error', 'message': 'File must contains at least two rows.'})
        return attributes, dimensions, data, csv_read_error 
    
    # check for at least one column
    if df.shape[1] < 1:
        csv_read_error.append({'type': 'error', 'message': 'File must have at least one column.'})
        return attributes, dimensions, data, csv_read_error

    # parse attributes, dimensions of data frame
    dimensions = [dict(name="row", type="int64", begin=0, end=len(df.index))]
    attributes = [dict(name=header, type="float64" 
        if df[header].dtype != "object" else "string") 
        for header in df.columns]
    
    # parse data
    data = []
    for header in df.columns.values:
        if df[header].dtype == "object":
            data.append(df[header].values.astype('unicode'))
        else:
            data.append(df[header].values.astype(float))
            
    # check for empty headers (pandas replaced them with 'Unnamed: <Column #>')
    empty_headers = []
    headers = df.columns.values
    for i in range(len(headers)):
        if headers[i].startswith('Unnamed:'):
            empty_headers.append(int(headers[i][8:])+1)

            # rename header so index starts at 1
            df = df.rename(columns = {headers[i]: "Unnamed: " + str(empty_headers[-1])})
    headers = df.columns.values

    # slycat warning for empty headers
    if len(empty_headers) != 0:
        csv_read_error.append({'type': 'warning', 'message': 'Found empty headers in columns: ' + \
                               str(empty_headers) + '.\nUsing "Unnamed: <Column #>" for empty headers.'})
    
    # look for duplicate headers (pandas adds .# to column)
    duplicated_headers = []
    for i in range(len(headers)):
        for j in range(i+1,len(headers)):
            if headers[j].startswith(headers[i]):
                header_j_suffix = headers[j].split('.')[-1]
                if header_j_suffix.isnumeric():
                    duplicated_headers.append(i+1)
                    duplicated_headers.append(j+1)
    duplicated_headers = pd.unique(duplicated_headers)

    # slycat warning for duplicated headers
    if len(duplicated_headers) != 0:
        csv_read_error.append({'type': 'warning', 'message': 'Found duplicated headers in columns: ' + \
                              str(duplicated_headers) + '.\nUsing "<Header>.#" for duplicate headers.'})

    # slycat warning that NaNs were discovered
    if df.isnull().values.any():
        csv_read_error.append({'type': 'warning', 'message': 'Found NaNs in .csv table.  Pandas converts a variety of null ' + 
                                "values to NaNs, including blanks, null, None, n/a, and others (see Panda's read_csv "
                                'for details).  NaNs may be ignored by Slycat algorithms/visualization tools.'})
        
    # headers may have been changed, need to recompute
    attributes = [dict(name=header, type="float64" 
        if df[header].dtype != "object" else "string") 
        for header in df.columns]
    
    # return data and errors
    return attributes, dimensions, data, csv_read_error



# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC . Under the terms of Contract
# DE-NA0003525 with National Technology and Engineering Solutions of Sandia, LLC, the U.S. Government
# retains certain rights in this software.





#!/bin/env python
# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC . Under the terms of Contract
# DE-NA0003525 with National Technology and Engineering Solutions of Sandia, LLC, the U.S. Government
# retains certain rights in this software.

"""Uploads a CSV file to Slycat Web Server to compute a CCA model."""

import slycat.web.client
import os

# slycat csv parser
import slycat.pandas_util

# parse arguments
def parser():

  parser = slycat.web.client.ArgumentParser(description="Create CCA model from .csv file.")

  # CSV file is mandatory
  parser.add_argument("file", default="-", help="Input CSV file.  Use - for stdin.  " +
                      "Default: %(default)s")
  
  # project/model name/description
  parser.add_argument("--project-description", default="", 
                      help="New project description.  Default: %(default)s")
  parser.add_argument("--project-name", default="CCA Models", 
                      help="New project name.  Default: %(default)s")
  parser.add_argument("--model-name", default="CCA from CSV", help="New model name.  " + 
                      "Default: %(default)s")
  parser.add_argument("--model-description", default="", help="New model description.  " +
                      "Default: %(default)s")
  
  # input/output columns for CCA
  parser.add_argument("--input-columns", default=[], nargs="+", help="Input column(s).")
  parser.add_argument("--output-columns", default=[], nargs="+", help="Output column(s).")

  parser.add_argument("--marking", default="mna", help="Marking type.  Default: %(default)s")

  # CCA algorithm scaling
  parser.add_argument("--scale-inputs", default=False, action="store_true", 
                      help="Enable input scaling.")

  parser.add_argument("--no-join", default=False, action="store_true", 
                      help="Don't wait for the model to finish.")
  
  return parser

# logging is just printing to the screen
def log (msg):
    print(msg)

# upload the model to the Slycat web server
def upload_model(arguments, attributes, dimensions, data, inputs, outputs, log):

  # get column names
  column_names = [attribute["name"] for attribute in attributes]
  
  # setup a connection to the Slycat Web Server.
  connection = slycat.web.client.connect(arguments)

  # Create a new project to contain our model.
  pid = connection.find_or_create_project(arguments.project_name, arguments.project_description)

  # Create the new, empty model.
  mid = connection.post_project_models(pid, "cca", arguments.model_name, arguments.marking, 
                                       arguments.model_description)

  # Upload our observations as "data-table".
  connection.put_model_arrayset(mid, "data-table")

  # Start our single "data-table" array.
  connection.put_model_arrayset_array(mid, "data-table", 0, dimensions, attributes)

  # Upload data into the array.
  for index, column in enumerate(data):
    log("Uploading column {} of {} ({})".format(index, len(data), column_names[index]))
    connection.put_model_arrayset_data(mid, "data-table", "0/%s/..." % index, [column])

  # Store the remaining parameters.
  connection.put_model_parameter(mid, "input-columns", inputs)
  connection.put_model_parameter(mid, "output-columns", outputs)
  connection.put_model_parameter(mid, "scale-inputs", arguments.scale_inputs)

  # Signal that we're done uploading data to the model.  This lets Slycat Web
  # Server know that it can start computation.
  connection.post_model_finish(mid)

  # Wait until the model is ready.
  connection.join_model(mid)

  return mid

# create CCA model
def create_model(arguments, log):

  # check that input file exists
  if not os.path.isfile(arguments.file):
    log("Input file " + arguments.file + " does not exist.")
    raise ValueError("Input file " + arguments.file + " does not exist.")

  # check that input/output columns are non-intersecting
  if set(arguments.input_columns).intersection(set(arguments.output_columns)):
    log("Input columns and output columns are intersecting.")
    raise ValueError("Input columns and output columns are intersecting.")

  # parse file using standard slycat csv parser
  attributes, dimensions, data, csv_read_error = \
    slycat.pandas_util.parse_file(arguments.file, file_name=True)

  # check for warnings/errors
  for i in range(len(csv_read_error)):
    if csv_read_error[i]["type"] == "warning":
      log("Warning: " + csv_read_error[i]["message"])
    else:
      log("Error: " + csv_read_error[i]["message"])
      raise ValueError(csv_read_error[i]["message"])

  # get column names/types
  column_names = [attribute["name"] for attribute in attributes]
  column_types = [attribute["type"] for attribute in attributes]
  columns = data

  # check that non-empty inputs are in headers
  if len(arguments.input_columns) == 0:
    log('You must specify at least one input column.')
    raise ValueError('You must specify at least one input column.  Available columns: %s' 
                     % str(column_names))
  else:
    for input_col in arguments.input_columns:
      if input_col not in column_names:
        log('Input column "' + input_col + '" not found in input file.')
        raise ValueError('Input column "' + input_col + '" not found in input file.')

  # check that non-empty outputs are in headers
  if len(arguments.output_columns) == 0:
    log('You must specifcy at least one output column.')
    raise ValueError('You must specify at least one output column.  Available columns: %s'
                     % str(column_names))
  else:
    for output_col in arguments.output_columns:
      if output_col not in column_names:
        log('Output column "' + output_col + '" not found in input file.')
        raise ValueError('Output column "' + output_col + '" not found in input file.')

  # get column indices for input/output arguments
  inputs = [column_names.index(input) for input in arguments.input_columns]
  outputs = [column_names.index(output) for output in arguments.output_columns]

  # check that inputs/outputs are numeric
  for input in inputs:
    if column_types[input] != "float64":
      raise Exception("Cannot analyze non-numeric input: %s" % column_names[input])
  for output in outputs:
    if column_types[output] != "float64":
      raise Exception("Cannot analyze non-numeric output: %s" % column_names[output])

  # upload model
  mid = upload_model (arguments, attributes, dimensions, data, inputs, outputs, log)

  # supply direct link to model
  host = arguments.host
  if arguments.port:
      host = host + ":" + arguments.port
  url = host + "/models/" + mid
  log("Your new model is located at %s" % url)
  log('***** CCA Model Successfully Created *****')

  return url

# main entry point
def main():

    # set up argument parser
    ps_parser = parser()  

    # get user arguments
    arguments = ps_parser.parse_args()

    # check arguments and create model
    url = create_model(arguments, log)

# command line version
if __name__ == "__main__":

    main()



# This module contains the model UI defaults for creating a DAC model.

# S. Martin
# 10/24/22

# returns a JSON variable with UI defaults
def dac_model_defaults():

    # from dac-ui.pref file:
    # ----------------------
    return {

        # the step size for the alpha slider (varies from 0 to 1)
        "ALPHA_STEP": 0.001,

        # default width for the alpha sliders (in pixels)
        "ALPHA_SLIDER_WIDTH": 170,

        # default height of alpha buttons (in pixels)
        "ALPHA_BUTTONS_HEIGHT": 33,

        # number of points over which to stop animation
        "MAX_POINTS_ANIMATE": 2500,

        # border around scatter plot (fraction of 1)
        "SCATTER_BORDER": 0.025,

        # scatter button toolbar height
        "SCATTER_BUTTONS_HEIGHT": 37,

        # scatter plot colors (css/d3 named colors)
        "POINT_COLOR": 'whitesmoke',
        "POINT_SIZE": 5,
        "NO_SEL_COLOR": 'gray',
        "SELECTION_1_COLOR": 'red',
        "SELECTION_2_COLOR": 'blue',
        "COLOR_BY_LOW": 'white',
        "COLOR_BY_HIGH": 'dimgray',
        "OUTLINE_NO_SEL": 1,
        "OUTLINE_SEL": 2,

        # pixel adjustments for d3 time series plots
        "PLOTS_PULL_DOWN_HEIGHT": 38,
        "PADDING_TOP": 10,        # 10 (values when plot selectors were)
        "PADDING_BOTTOM": 14,     # 24 (at the bottom of the plots)
        "PADDING_LEFT": 37,
        "PADDING_RIGHT": 10,
        "X_LABEL_PADDING": 4,
        "Y_LABEL_PADDING": 13,
        "LABEL_OPACITY": 0.2,
        "X_TICK_FREQ": 80,
        "Y_TICK_FREQ": 40,

    }